/*
 *
 * MediaServerSession.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: MediaServerSession.cxx,v $
 * Revision 1.5  2007/07/14 07:50:32  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.4  2007/06/08 15:10:32  joegenbaclor
 * Added IVR DTMF Grammar
 *
 * Revision 1.3  2007/06/08 02:33:47  joegenbaclor
 * Commiting more Media Server Related Changes
 *
 * Revision 1.2  2007/06/05 16:27:56  joegenbaclor
 * Added IVR audio playback functions to B2BUAConnection
 *
 * Revision 1.1  2007/06/04 16:05:33  joegenbaclor
 * *** empty log message ***
 *
 *
 */


#include "MediaServerSession.h"
#include "B2BUAConnection.h"

using namespace MS;
using namespace B2BUA;

#define new PNEW


IVRContext::IVRContext(
  MediaServerSession * _conn,
  PTextToSpeech * tts,
  BOOL autoDelete
  ) : OpalVXMLSession( _conn, tts, autoDelete )
{
  m_Session = _conn;
  m_CurrentGrammar = NULL;
}

MediaServerSession * IVRContext::GetSession()const
{
  return m_Session;
}

BOOL IVRContext::Open(
  const PString & _mediaFormat
)
{
  mediaFormat = _mediaFormat;

  PVXMLChannel * chan = PFactory<PVXMLChannel>::CreateInstance(mediaFormat);
  if (chan == NULL) {
    PTRACE(1, "VXML\tCannot create VXML channel with format " << mediaFormat);
    return FALSE;
  }

  chan->SetVXMLSession( this );

  // set the underlying channel
  if (!PIndirectChannel::Open(chan, chan))
    return FALSE;

  // start the VXML session in another thread
  {
    PWaitAndSignal m(sessionMutex);
    if (!chan->Open(this))
      return FALSE;
    vxmlChannel = chan;
  }

  B2BUAConnection * conn = dynamic_cast<B2BUAConnection *>(m_B2BUAConnection.GetObject());
  if( conn != NULL )
    return conn->IVROnOpenChannel( this );

  return FALSE;
}

void IVRContext::FlushQueue()
{
  PWaitAndSignal lock(sessionMutex);
  vxmlChannel->FlushQueue();
}


BOOL IVRContext::Close()
{
  PWaitAndSignal lock( m_GrammarMutex );
  if( m_CurrentGrammar != NULL )
  {
    m_CurrentGrammar->StopCollecting();
    delete m_CurrentGrammar;
    m_CurrentGrammar = NULL;
  }

  return OpalVXMLSession::Close();
}

PWAVFile * IVRContext::CreateWAVFile(
  const PFilePath & fn,
  PFile::OpenMode mode,
  int opts,
  unsigned fmt
)
{
  return OpalVXMLSession::CreateWAVFile( fn, mode, opts, fmt );
}

BOOL IVRContext::AttachB2BUAConnection(
  SIPSession * session
)
{
  if( session != NULL )
    m_B2BUAConnection = GCCREATEREF( session, "IVRContext::AttachB2BUAConnection", "B2BUAConnection" );

  return m_B2BUAConnection != NULL;
}

void IVRContext::OnEndPlay( const PString & identifier )
{
  B2BUAConnection * conn = dynamic_cast<B2BUAConnection *>(m_B2BUAConnection.GetObject());
  if( conn != NULL )
    conn->IVROnEndPlaying( identifier );
}

BOOL IVRContext::OnUserInput(const PString & str)
{
  B2BUAConnection * conn = dynamic_cast<B2BUAConnection *>(m_B2BUAConnection.GetObject());
  
  conn->IVROnUserInput( str );

  {
    PWaitAndSignal lock( m_GrammarMutex );
    if( m_CurrentGrammar != NULL )
    {
      for( PINDEX i = 0; i < str.GetLength(); i++ )
        m_CurrentGrammar->OnDTMFInput( str[i] );
    }
  }  

  return TRUE;
}

BOOL IVRContext::SetDTMFGrammar(
  const PString & identifier,
  int maxDigits, /// maximum digitd collected
  char termChar, /// terminating char to end collection
  BOOL canStopPlayBack,
  const PTimeInterval & collectTimeout,
  const PTimeInterval & perDigitTimeout
)
{
  PWaitAndSignal lock( m_GrammarMutex );
  if( m_CurrentGrammar != NULL )
  {
    m_CurrentGrammar->StopCollecting();
    delete m_CurrentGrammar;
    m_CurrentGrammar = NULL;
  }

  m_CurrentGrammar = new IVRDTMFGrammar( 
    identifier,
    this,
    maxDigits,
    termChar,
    canStopPlayBack,
    collectTimeout,
    perDigitTimeout );

  return TRUE;
}

void IVRContext::OnEndCollection(
  const PString & grammarId,
  const PString & grammarBuffer
)
{
  B2BUAConnection * conn = dynamic_cast<B2BUAConnection *>(m_B2BUAConnection.GetObject());
  if( conn != NULL )
    conn->IVROnEndCollection( grammarId, grammarBuffer );
}

//////////////////////////////////////////////////////////////////////////

MediaServerSession::MediaServerSession(
  OpalCall & call,            ///<  Owner calll for connection
  MediaServerEndPoint & endpoint, ///<  Owner endpoint for connection
  const PString & token      ///<  Token for connection
  ) : OpalIVRConnection( call, endpoint, token, NULL, OString::Empty() )
{
}

/////////////////////////////////////////////////////////////////


IVRDTMFGrammar::IVRDTMFGrammar(
  const PString & identifier,
  IVRContext * context,
  int maxDigits, /// maximum digitd collected
  char termChar, /// terminating char to end collection
  BOOL canStopPlayBack,
  const PTimeInterval & collectTimeout,
  const PTimeInterval & perDigitTimeout
)
{
  m_Identifier = identifier;
  m_CanCollectDigits = TRUE;
  m_IVRContext = context;
  m_MaxDigits = maxDigits;
  m_TermChar = termChar;
  m_CanStopPlayback = canStopPlayBack;

  m_CollectTimer = PTimer( collectTimeout );
  m_CollectTimer.SetNotifier( PCREATE_NOTIFIER( OnCollectTimeout ) );
  m_CollectTimer.Resume();

  m_DigitTimer = PTimer( perDigitTimeout );
  m_DigitTimer.SetNotifier( PCREATE_NOTIFIER( OnDigitTimeout ) );
  m_DigitTimer.Resume();

}

void IVRDTMFGrammar::OnDTMFInput( 
  char dtmf
)
{
  PWaitAndSignal lock( m_DigitsMutex );
  if( !m_CanCollectDigits )
    return;
  
  if( dtmf != m_TermChar )
  {
    m_Collected += dtmf;
    if( m_CanStopPlayback )
    {
      m_IVRContext->FlushQueue();
      m_CanStopPlayback=FALSE;
    }
    if( m_Collected.GetLength() >= m_MaxDigits )
      OnEndCollection();
  }else
  {
    OnEndCollection();
  }
}

void IVRDTMFGrammar::OnEndCollection()
{
  PWaitAndSignal lock( m_DigitsMutex );
  StopCollecting();

  m_IVRContext->OnEndCollection( m_Identifier, m_Collected );
}

void IVRDTMFGrammar::OnCollectTimeout( PTimer & timer, INT data )
{
  PWaitAndSignal lock( m_DigitsMutex );
  OnEndCollection();
}

void IVRDTMFGrammar::OnDigitTimeout( PTimer & timer, INT data )
{
  PWaitAndSignal lock( m_DigitsMutex );
  OnEndCollection();
}

void IVRDTMFGrammar::StopCollecting()
{
  PWaitAndSignal lock( m_DigitsMutex );
  m_CanCollectDigits = FALSE;
  m_CollectTimer.Stop();
  m_DigitTimer.Stop();
}



