/*
 *
 * SIPHeaderB.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: SIPHeaderB.cxx,v $
 * Revision 1.11  2007/07/14 07:50:33  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.10  2007/06/26 02:22:53  joegenbaclor
 * Bug fix:  SIPHeaderB operator=() does not call MakeUnique on parameters.   The effect
 *  of this is when SIPSessions send responses with a to tag, the original INVITE would
 *  inherit the tag automatically.   This might be a desired behavior but might lead
 *  to confusion as how the original INVITE got the tag.
 *
 * B2BUACall And CallSession are updated accordingly to comply with this fix
 *
 * Revision 1.9  2007/03/02 06:22:46  joegenbaclor
 * Corrected some bug in GC.
 * Added CPU Throttle mechanism during heavy load to avoid denial of service
 * Made sure lists are Made unique in SIPHeade clone function
 *
 * Revision 1.8  2006/11/30 08:19:38  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.7  2006/05/19 06:30:37  joegenbaclor
 * 1.  Fixed bug in SIPHeaderB where tag may not be parsed properly if no < > enclosure is present
 * 2.  Various enhancements to OpenSBC and proxy session
 *
 * Revision 1.6  2006/05/17 03:43:07  joegenbaclor
 * Bulk commit after sourceforge cvs migration changes
 *
 * Revision 1.5  2006/03/21 14:11:51  joegenbaclor
 * 1.  Added rport support for NAT Traversal
 * 2.  Added NAT Keep alive mechanism in SIPSession
 * 3.  Some Improvements in various classes affecting NAT like Via, SIPTransport etc.
 *
 * Revision 1.4  2006/03/17 13:51:00  joegenbaclor
 * 1.  Implemented UAC Call Redirection
 * 2.  Implemented SendCancel() method in CallSession
 * 3.  Introduced ClearInternalHeaders() function in SIP Message to erase internal headers inheritted from previous transaction.
 *
 * Revision 1.3  2006/03/14 03:53:53  joegenbaclor
 * Initial source upload for BETA Release Candidate
 *
 *
 *
 */

#include "SIPHeaderB.h"

#define new PNEW

using namespace SIPParser;

SIPHeaderB::SIPHeaderB( 
  const OString & headerName
  ) : SIPHeader( headerName )
{
  m_HasURIEnclosure = FALSE;
}

SIPHeaderB::SIPHeaderB(
  const OString & headerName,
  const SIPHeaderB & header 
)  : SIPHeader( headerName )
{
  m_HasURIEnclosure = FALSE;
  operator=( header );
}

SIPHeaderB::SIPHeaderB(
  const OString & headerName,
  const OString & header 
)  : SIPHeader( headerName )
{
  m_HasURIEnclosure = FALSE;
  operator=( header );
}

SIPHeaderB::SIPHeaderB(
  const OString & headerName,
  const MimeHeader & header
)  : SIPHeader( headerName )
{
  m_HasURIEnclosure = FALSE;
  m_HeaderBody = header.GetMimeValue();
  operator=( header );
}

SIPHeaderB::SIPHeaderB(
  const OString & headerName,
  const SIPURI & header,
  const OString & displayName,
  const OString & tag
) : SIPHeader( headerName )
{
  m_HasURIEnclosure = FALSE;
  m_URI = header;
  m_DisplayName = displayName;
  
  if( !tag.IsEmpty() )
    AddParameter( "tag", tag, TRUE );
}

SIPHeaderB & SIPHeaderB::operator=( 
  const SIPHeaderB & header 
)
{
  m_DisplayName = header.m_DisplayName;
  m_HeaderName = header.m_HeaderName;
  m_HeaderBody = header.m_HeaderBody;
  m_Parameters = header.m_Parameters;
  m_Parameters.MakeUnique();
  m_URI = header.m_URI;
  m_HasURIEnclosure = header.m_HasURIEnclosure;
  return *this;
}

SIPHeaderB & SIPHeaderB::operator=(
  const OString & header 
)
{
  MimeHeader h( header );
  return operator=( h );
}

SIPHeaderB & SIPHeaderB::operator=(
  const MimeHeader & header
)
{
  OString mimeValue = header.GetMimeValue();
  /// SIPHeaderB: "Bob" <sips:bob@biloxi.com> ;tag=a48s
  /// SIPHeaderB: sip:+12125551212@phone2net.com;tag=887s
  /// SIPHeaderB: Anonymous <sip:c8oqz84zk7z@privacy.org;transport=udp>;tag=hyh8
   
  PINDEX hasDisplayName = mimeValue.Find( "<" );
  PINDEX greater = mimeValue.Find( ">" );
  PINDEX hasParams = mimeValue.Find( ";", greater );
  OString uriString = mimeValue;
  if( hasDisplayName != P_MAX_INDEX )
  {
    /// parse the display name
    m_HasURIEnclosure = TRUE;
    m_DisplayName = mimeValue.Left( hasDisplayName ).Trim();
    uriString = uriString.Mid( hasDisplayName + 1, greater - hasDisplayName -1 );
  }

  if( hasParams != P_MAX_INDEX )
  {
    /// parse the parameters
    OString params = mimeValue.Mid( hasParams + 1 );
    MimeParam::CreateCollection( params, m_Parameters );
    OString uriString = mimeValue.Left( hasParams );
  }

  m_URI = uriString;
  
  return *this;
}

PObject * SIPHeaderB::Clone()const
{
  SIPHeaderB * h = new SIPHeaderB( m_HeaderName );
  h->m_DisplayName = m_DisplayName;
  h->m_HeaderName = m_HeaderName;
  h->m_HeaderBody = m_HeaderBody;
  h->m_Parameters = m_Parameters;
  h->m_Parameters.MakeUnique();
  h->m_URI = m_URI;
  h->m_HasURIEnclosure = m_HasURIEnclosure;
  return h;
}

void SIPHeaderB::PrintOn(
  ostream & strm
)const
{
  if( !GetHeaderName().IsEmpty() )
    strm << GetHeaderName() << ": ";
  
  BOOL hasDisplayName = !m_DisplayName.IsEmpty();

  if( hasDisplayName )
  {
    ///if( m_DisplayName[0] != '\"' )
    ///  strm << "\"" << m_DisplayName << "\"";
    ///else
    strm << m_DisplayName;

    strm << " <" << m_URI << ">";
    
  }else
  {
    if( m_Parameters.GetSize() > 0 || m_HasURIEnclosure )
      strm << " <"; 
    
    strm << m_URI;
    
    if( m_Parameters.GetSize() > 0 || m_HasURIEnclosure )
      strm << ">";
  }

  if( m_Parameters.GetSize() > 0 )
  {
    for( PINDEX i = 0; i < m_Parameters.GetSize(); i++ )
      strm << ";" << m_Parameters.GetDataAt( i );
  }

}

void SIPHeaderB::PrepareBody()
{
  OStringStream strm;

  BOOL hasDisplayName = !m_DisplayName.IsEmpty();

  if( hasDisplayName )
  {
    strm << m_DisplayName;
    strm << " <" << m_URI << ">";
  }else
  {
    strm << m_URI;
  }

  if( m_Parameters.GetSize() > 0 )
  {
    for( PINDEX i = 0; i < m_Parameters.GetSize(); i++ )
      strm << ";" << m_Parameters.GetDataAt( i );
  }

  m_HeaderBody = strm.str();
}

void SIPHeaderB::AddParameter(
  const OString & name,
  const OString & value,
  BOOL replaceIfExist
)
{

  BOOL exist = m_Parameters.Contains( name.c_str() );
  if( !exist )
  {
    m_Parameters.SetAt( name.c_str(), new MimeParam( name, value ) );
  }else
  {
    if( replaceIfExist )
    {
      MimeParam * param = m_Parameters.GetAt( name.c_str() );
      //PAssertNULL( param );
      param->SetParamName( name );
      param->SetParamValue( value );
    }
  }
}

void SIPHeaderB::SetParameter(
  const OString & name,
  const OString & value
)
{
  BOOL exist = m_Parameters.Contains( name.c_str() );
  if( exist )
  {
      MimeParam * param = m_Parameters.GetAt( name.c_str() );
      //PAssertNULL( param );
      param->SetParamName( name );
      param->SetParamValue( value );
  }else
  {
    AddParameter( name, value );
  }
}

PINDEX SIPHeaderB::FindParameter(
  const OString & paramName
)
{
  for( PINDEX i = 0; i < m_Parameters.GetSize(); i++ )
  {
    if( m_Parameters.GetDataAt( i ).GetParamName() *= paramName )
      return i;
  }

  return P_MAX_INDEX;
}

BOOL SIPHeaderB::GetParameter(
  const OString & name,
  OString & value
)const
{
  MimeParam * param = m_Parameters.GetAt( name.c_str() );
  
  if( param != NULL )
  {
    value = param->GetParamValue();
    return TRUE;
  }

  return FALSE;
}

MimeParam & SIPHeaderB::GetParameter(
  PINDEX index
)const
{
  return m_Parameters.GetDataAt(index);
}

