/*
 *
 * SIPTransaction.cxx
 * 
 * Open SIP Stack ( OSS )
 *
 * Copyright (C) 2006 Joegen E. Baclor. All Rights Reserved.
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is OpenSIPStack Library.
 *
 * The Initial Developer of the Original Code is Joegen E. Baclor.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * $Log: SIPTransaction.cxx,v $
 * Revision 1.40  2007/07/15 12:33:21  joegenbaclor
 * Corrected bug in calling OnIncomingSDPOffer where it is called for the second leg instead of the first leg call session.  This is due to the change in placement of the method form receipt of the invite to SetupOutbound
 *
 * Revision 1.39  2007/07/14 07:50:33  joegenbaclor
 * Migrated code to use new STL based string class
 *
 * Revision 1.38  2007/07/03 17:17:37  joegenbaclor
 * Got rid of compile warnings in unix builds
 *
 * Revision 1.37  2007/06/13 14:13:34  joegenbaclor
 * Fixed bug in DoDNSFailover where the attempt count is not propagated to the new transaction
 *
 * Revision 1.36  2007/05/25 08:38:09  joegenbaclor
 * Introduced GCVERIFYREF and GCCREATEREF Macro
 * Transfered B2BUAConnection::OnIncomingCall to be called in B2BUAEndPoint::OnIncomingSIPMessage
 *
 * Revision 1.35  2007/05/13 03:17:58  joegenbaclor
 * Added DNS Failover support
 *
 * Revision 1.34  2007/05/08 12:49:28  joegenbaclor
 * Corrected memory leaks in SIPMessage.
 *
 * Revision 1.33  2007/04/10 06:13:08  joegenbaclor
 * Added new Encryption Mode
 *
 * Revision 1.32  2007/03/28 09:09:24  joegenbaclor
 * Logging enhancements
 *
 * Revision 1.31  2007/03/21 05:30:44  joegenbaclor
 * Logging aesthetics
 *
 * Revision 1.30  2007/02/06 16:50:05  joegenbaclor
 * 1.  More work on TCP
 * 2.  Mutexed SIPHeadeF and SIPHeaderC collection members
 * 3.  Correct bug in B2BUACall record routing
 * 4.  Added Strict Router support to SIPSession::SendRequest
 *
 * Revision 1.29  2007/02/05 03:55:06  rcolobong
 * Fix compilation problem in ProcessEvent where m_Socket is not defined if ENABLE_TCP_TRANSPORT is disable
 *
 * Revision 1.28  2007/02/04 15:04:10  joegenbaclor
 * Completed transaction layer TCP support
 *
 * Revision 1.27  2007/01/29 05:59:08  joegenbaclor
 * ENABLE_TCP_TRANSPORT macro around TCP specific code
 *
 * Revision 1.26  2007/01/29 02:49:16  joegenbaclor
 * Sanity checks for sending request over ICT
 *
 * Revision 1.25  2007/01/27 23:06:51  joegenbaclor
 * More work on TCP support
 *
 * Revision 1.24  2007/01/26 06:51:23  joegenbaclor
 * 1.  More work on TCP support
 * 2.  Fixed bug where CANCEL is not sent if 180 is not received yet
 *
 * Revision 1.23  2006/11/30 08:19:38  joegenbaclor
 * Relicensed opensipstack to MPL/GPL/LGPL triple license scheme
 *
 * Revision 1.22  2006/11/21 06:55:19  joegenbaclor
 * Added ability to add/append additional listeners.
 *
 * Revision 1.21  2006/11/08 06:46:21  joegenbaclor
 * Added session level NIC/Interface binding
 *
 * Revision 1.20  2006/11/07 09:30:46  joegenbaclor
 * Added new internal header "IFACE-ADDR" to allow transctions to remember the interface address for sending future responses within the transaction lifespan
 *
 * Revision 1.19  2006/09/06 07:36:38  joegenbaclor
 * Revised garbage collector so that each Garbage collectible objects will have its own
 *  instance of the GC.  This eliminates the need to cast to PObject
 *
 * Revision 1.18  2006/09/04 02:58:53  joegenbaclor
 * 1.  More bug fixes to garbage collector
 * 2.  Added new Logging macro so call System Log for fatal errors
 *
 * Revision 1.17  2006/09/01 03:12:26  joegenbaclor
 * Bulk Commit:
 *
 * 1.  A bug is discovered in SIP Transaction Destructor where destruction of the SIPTimers may cause a deadlock on the timer manager effectivily freezing the entire message Queue System.  This was fixed by introducing a separate function to destroy the timers on receipt of the final event just before the transaction is queued for deletion.
 *
 * 2.  There is bug in Invite Server Transaction where ACK transaction is added to the Ack Transaction Pool after sending og none 2xx final response.  This new transaction is never removed from the pool after Timer_H has fired.  this may result to a bad pointer if ACK is received after the timer expired since the transaction may have been deleted by then.
 *
 * 3.  A few tweaks in GC GetName func to return the Class Name if GC NAme is empty.
 *
 * Revision 1.16  2006/08/29 03:56:44  joegenbaclor
 * 1.  Added GC Reference in SIP Event
 * 2.  Corrected buf in ProcessStackEvent where session is checked against NULL not allowing requests to create new sessions
 *
 * Revision 1.15  2006/08/28 07:18:35  joegenbaclor
 * Bulk Commit: Logging overhaul and Call Session callback changes to reflect SIPMessage eg:  OnAlerting(), OnProgress(), etc
 *
 * Revision 1.14  2006/08/14 01:29:09  joegenbaclor
 * Introduced separate provisions for UAS/UAC Current Request/Reponse.  Previousely the current request/response can be accessed via a single provision SIPSession::Get/Set/CurrentRequest()
 *
 * Revision 1.13  2006/04/18 07:18:21  joegenbaclor
 * Fixed unix build environment errors and warnings
 *
 * Revision 1.12  2006/03/25 18:40:21  joegenbaclor
 * 1.  Worked on some race conditions on Finite State Machines during heavy load.
 * 2.  Intoduced Global Mutex to SIPMessage.  This is inevitable beacuse PDictionary and PList members are not thread-safe
 * 3,  Some minor improvements to simple-proxy application
 *
 * Revision 1.11  2006/03/24 15:42:18  joegenbaclor
 * Some bug fixes including possible deadlocks between Session Destruction and ProcessStackEvent
 *
 * Revision 1.10  2006/03/22 08:45:52  joegenbaclor
 * More work on simple-proxy and affected classes
 *
 * Revision 1.9  2006/03/17 13:51:00  joegenbaclor
 * 1.  Implemented UAC Call Redirection
 * 2.  Implemented SendCancel() method in CallSession
 * 3.  Introduced ClearInternalHeaders() function in SIP Message to erase internal headers inheritted from previous transaction.
 *
 * Revision 1.8  2006/03/14 03:53:53  joegenbaclor
 * Initial source upload for BETA Release Candidate
 *
 *
 *
 */

#include "SIPTransaction.h"
#include "SIPTransactionStage.h"
#include "SIPTransactionManager.h"
#include "ossbuildopts.h"
#include "SIPTransport.h"

#define new PNEW

using namespace SIPTransactions;
using namespace SIPTransports;

PInt64 SIPTransaction::m_InternalIdCounter = 0;

#if LOG_INSTANCE_COUNTER
PAtomicInteger SIPTransaction::m_InstanceCounter = 0;
#endif

GCCREATECOLLECTOR( SIPTransaction );

SIPTransaction::SIPTransaction(
  const TransactionId & id,
  Type type,
  SIPTransactionManager & manager
) : SIPTimerInterface( manager ),
    m_Type( type ),
    m_Id( id ),
    m_State( 0 ),
    m_IsTerminating( FALSE ),
    m_IsTerminated( FALSE ),
    m_Manager( manager )
{
#if LOG_INSTANCE_COUNTER
  ++SIPTransaction::m_InstanceCounter;
#endif

#if ENABLE_TCP_TRANSPORT
  m_Socket = NULL;
  m_AutoDeleteSocket = FALSE;
  m_TransportThread = NULL;
#endif
  m_IsReliableTransport = FALSE;
  m_InterfacePort = 0;

  m_FailOverAttempts = 0;

  m_CallId = id.GetCallId();

  if( SIPTransaction::m_InternalIdCounter == 0 )
    SIPTransaction::m_InternalIdCounter = PTimer::Tick().GetMilliSeconds();

  ++SIPTransaction::m_InternalIdCounter;
  ///if( ++SIPTransaction::m_InternalIdCounter == INT_MAX )
  ///  SIPTransaction::m_InternalIdCounter = 0;

  switch( type )
  {
    case ICT:
      m_InternalId = "ICT(" + OString(SIPTransaction::m_InternalIdCounter) + ")";
      break;
    case NICT:
      m_InternalId = "NICT(" + OString(SIPTransaction::m_InternalIdCounter) + ")";
      break;
    case IST:
      m_InternalId = "IST(" + OString(SIPTransaction::m_InternalIdCounter) + ")";
      break;
    case NIST:
      m_InternalId = "NIST(" + OString(SIPTransaction::m_InternalIdCounter) + ")";
      break;
  }

  ((SIPTimerInterface*)this)->m_InternalId = m_InternalId;

  StartTransactionLifeSpanTimer( SIP_MAX_TRANSACTION_LIFE_SPAN );

  SetLogContextId( id.GetCallId() );
  m_TimerInterfaceLogger = dynamic_cast<Logger&>(*this);

  LOG_IF_DEBUG( LogDetail(), m_InternalId << "\t" << "*** CREATED *** - " << m_Id );
}

SIPTransaction::~SIPTransaction()
{
#if LOG_INSTANCE_COUNTER
  --SIPTransaction::m_InstanceCounter;
#endif

  LOG_IF_DEBUG( LogDetail(), m_InternalId << "\t" << "*** DESTROYED *** - " << m_Id );
}

void SIPTransaction::EnqueueEvent(
  SIPEvent * event
)
{
  PWaitAndSignal lock( m_EventQueueMutex );
  if( m_IsTerminating && !PIsDescendant( event, SIPFinalEvent ) )
  {
    delete event;
    return;
  }

  SIPTransaction::GCRef me = GCCREATEREF( this, "SIPTransaction::EnqueueEvent", "SIPTransaction" );
  if( me == NULL )
  {
    delete event;
    return;
  }

  event->m_GCRef = new SIPTransaction::GCRef( me );
  
  m_EventQueue.Enqueue( event );
  m_Manager.m_TransactionStage->Enqueue( this );
 
}

SIPEvent * SIPTransaction::DequeueEvent()
{
  PWaitAndSignal lock( m_EventQueueMutex );
  return m_EventQueue.Dequeue();
}

void SIPTransaction::FlushEvents()
{
  PWaitAndSignal lock( m_InternalLock );

  for( ;; )
  {
    SIPEvent * event = DequeueEvent();
    if( event == NULL )
      break;
    delete event;
  }
}

void SIPTransaction::ProcessEvents()
{

    PWaitAndSignal lock( m_InternalLock );

    SIPTransaction::GCRef autoRef = GCCREATEREF( this, "SIPTransaction::ProcessEvents", "SIPTransaction" );
    if( autoRef == NULL )
      return;

    SIPEvent * event = DequeueEvent();

    if( event == NULL )
      return;

    if( m_IsTerminated )
    {
      delete event;
      return;
    }
    
    if( PIsDescendant( event->GetEvent(), SIPTimerEvent ) )
    {
      LOG_IF_DEBUG( LogDetail(), m_InternalId << "\t" << (*(event->GetEvent())) );
      OnTimerEvent( (SIPTimerEvent&)(*(event->GetEvent())) );
    }else if( PIsDescendant( event->GetEvent(), SIPMessage ) )
    {
      SIPMessage & msg = (SIPMessage&)(*(event->GetEvent()));
      StartLine startLine;
      msg.GetStartLine( startLine );
      
      OString sendStatus = msg.GetInternalHeader( "SEND-EVENT" );
      BOOL isSendEvent = sendStatus *= "TRUE";

#if ENABLE_TCP_TRANSPORT
      m_AutoDeleteSocket = msg.CanTransactionDeleteSocket();
      if( !isSendEvent && msg.IsRequest() && 
        m_Socket == NULL && 
        msg.GetSocket() != NULL /*&&
        ( GetType() == IST || GetType() == NIST )*/ )
      {
        StartReadingFromSocket( msg.GetSocket(), msg.CanTransactionDeleteSocket() );
      }
#endif

      LOG_IF_DEBUG( LogDetail(), m_InternalId << "\tEvent(SIPMessage) - " << startLine );
      OnSIPEvent( (SIPMessage&)(*(event->GetEvent())), isSendEvent );

    }else if( PIsDescendant( event, SIPFinalEvent ) )
    {
      LOG_IF_DEBUG( LogDetail(), m_InternalId << "\t" << (*(event->GetEvent())) );
      OnFinalEvent();
      m_Manager.m_TransactionStage->TerminateTransaction( this );
      m_IsTerminated = TRUE;

    }else if( PIsDescendant( event, SIPCancelEvent )  )
    {
      LOG_IF_DEBUG( LogDetail(), m_InternalId << "\t" << (*(event->GetEvent())) );
      OnCancelEvent();
    }
    else
    {
      LOG_IF_DEBUG( LogDetail(), m_InternalId << "\t" << (*(event->GetEvent())) );
      OnUnknownEvent( event );
      return;
    }

    delete event;

}

void SIPTransaction::OnTimerEvent( 
  SIPTimerEvent & 
)
{
}

void SIPTransaction::OnSIPEvent(
  SIPMessage &,
  BOOL
)
{
}

void SIPTransaction::OnFinalEvent()
{
  m_IsTerminating = TRUE;
  m_Manager.RemoveTransaction( GetIdentifier() );
  RemoveAllTimers();
  FlushEvents();

#if ENABLE_TCP_TRANSPORT
  if( m_TransportThread != NULL )
  {
    PAssertNULL( m_Socket );
    m_Socket->Close();
    delete m_TransportThread;
    m_TransportThread = NULL;
    if( m_AutoDeleteSocket )
    {
      delete m_Socket;
      m_Socket = NULL;
    }
  }
#endif

}

void SIPTransaction::OnCancelEvent()
{
}

void SIPTransaction::OnUnknownEvent( 
  SIPEvent * event 
)
{
  delete event;
}

void SIPTransaction::OnTransactionLifeSpanExpire(
  SIPTimer & timer
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::LifeSpanExpire );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_A( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::A );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_B( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::B );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_C( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::C );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_D( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::D );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_E( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::E );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_F( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::F );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_G( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::G );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_H( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::H );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_I( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::I );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_J( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::J );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::OnTimerExpire_K( 
  SIPTimer & timer 
)
{
  SIPTimerEvent * event = new SIPTimerEvent( SIPTimerEvent::K );
  event->SetInterval( timer.GetInterval() );
  EnqueueEvent( new SIPEvent( event ) );
}

void SIPTransaction::Cleanup()
{
}

void SIPTransaction::Terminate()
{
  m_IsTerminating = TRUE;
  EnqueueEvent( new SIPFinalEvent() );
}

void SIPTransaction::SetOpeningRequest( const SIPMessage & message )
{ 
  m_OpeningRequest = message;
  m_InterfaceAddress = message.GetInterfaceAddress();
  m_InterfacePort = message.GetInterfacePort();
}


/////////////////////////////////////////////////////////////////////

class TransportThread : public PThread
{
  PCLASSINFO( TransportThread, PThread );
public:

  TransportThread(
    SIPTransaction * trn 
   ) : PThread( 1024, NoAutoDeleteThread, NormalPriority,  "TransportThread" )
  {
    m_Transaction = trn;
    m_IsTerminating = FALSE;
    Resume();
  }

  ~TransportThread()
  {
    m_IsTerminating = TRUE;
    WaitForTermination();
  }

  void Main()
  {
    //SIPTransport::NotifyRead( "TCP Read Thread STARTED\r\n" );
    for(;;)
    {
      if( m_IsTerminating )
        break;
      if( !m_Transaction->ProcessRead() )
        break;
    }

    m_Transaction->OnSocketReadTerminated();

    //SIPTransport::NotifyRead( "TCP Read Thread TERMIANTED\r\n" );

  }

  BOOL m_IsTerminating;
  SIPTransaction * m_Transaction;
};

/////////////////////////////////////////////////////////////////////


void SIPTransaction::StartReadingFromSocket(
  PSocket * socket,
  BOOL autoDelete
)
{
#if ENABLE_TCP_TRANSPORT
  m_IsReliableTransport = TRUE;
  if( m_TransportThread == NULL )
  {
    m_Socket = socket;
    m_AutoDeleteSocket = autoDelete;
    m_TransportThread = new TransportThread( this );
  }
#endif
}

BOOL SIPTransaction::ProcessRead()
{
#if ENABLE_TCP_TRANSPORT
  PAssertNULL( m_Socket );
  SIPMessage * msg = new SIPMessage();
  if( msg->Read( *m_Socket ) )
  {
    OString packet = msg->AsString();
    PIPSocket::Address rcvAddr;
    WORD rcvPort;
    static_cast<PTCPSocket*>(m_Socket)->GetPeerAddress( rcvAddr, rcvPort );
    OStringStream traceStream;
    traceStream << "RCV: enc=" << msg->IsEncrypted() << " " 
      << packet.GetLength() << " Bytes from " << rcvAddr << ":" << rcvPort 
      << ":TCP (" << msg->GetStartLine() << ")" 
      << "\r\n" << packet;

    PTRACE( 5, traceStream.str() );
    SIPTransport::NotifyRead( traceStream.str() );

    msg->SetSocket( m_Socket );
    
    EnqueueEvent( new SIPEvent( msg ) );
    return TRUE;
  }else
  {
    delete msg;
    return FALSE;
  }
#endif
  return FALSE;
}

void SIPTransaction::OnSocketReadTerminated()
{
}

PTCPSocket * SIPTransaction::CreateTCPSocket(
  const SIPURI & targetAddress
)
{
#if ENABLE_TCP_TRANSPORT

  PIPSocket::Address addr;
  WORD port = 0;

  addr = targetAddress.GetAddress();
  if( !addr.IsValid() )
    return NULL;
  
  port = (WORD)targetAddress.GetPort().AsInteger();
  if( port == 0 )
    port = 5060;

  PTCPSocket * s = new PTCPSocket(port);
  if( !s->Connect(addr) )
  {
    delete s;
    return NULL;
  }

  StartReadingFromSocket( s, FALSE );

  return s;
#else
  return NULL;
#endif
}

PTCPSocket * SIPTransaction::CreateTCPSocket(
  SIPMessage & msg
)
{
#if ENABLE_TCP_TRANSPORT
  PAssert( msg.IsRequest(), "Sending a response via a Client Transaction is forbidden!" );
  SIPURI requestURI;
  if( msg.HasRoute() && msg.GetRouteSize() > 0 )
  {
    SIPParser::Route route = msg.GetRouteAt( 0 );
    if( !route.GetTopURI( requestURI ) )
      return FALSE;
  }else
  {
    
    if( !msg.GetRequestURI( requestURI ) )
      return FALSE;
  }

  requestURI.AddParameter( "transport", "TCP", TRUE );

  PTCPSocket * sock = CreateTCPSocket( requestURI );

  if( sock != NULL )
    m_SocketTargetAddress = requestURI;

  return sock;
#else
  return NULL;
#endif
}

BOOL SIPTransaction::SendMessageToTransport(
  SIPMessage & msg
)
{
  PWaitAndSignal lock( m_SendMutex );

  if( !msg.IsValid() || m_IsTerminating )
    return FALSE;

#if ENABLE_TCP_TRANSPORT

  OString packet = msg.AsString();
  int len = packet.GetSize();

  BOOL useTCP = FALSE;
  if( m_Socket == NULL && msg.IsRequest() )
  {
    SIPURI ruri;
    msg.GetRequestURI( ruri );
    useTCP = ruri.GetParameter( "transport" ) *= "tcp";
  }

  if( m_Socket == NULL && len <= 1300 && !useTCP ) /// let the transport manager handle it
  {
    m_Manager.OnSendMessageToTransport( msg, *this );
    return TRUE;
  }else 
  {
    if( m_Socket == NULL && ( m_Type == NICT || m_Type == ICT ) )
    {
      /// Uh-oh! We have a large message
      /// Lets create a new transport to accomodate it
      /// and adjust the via accordingly
      m_Socket = CreateTCPSocket( msg );
      if( m_Socket == NULL )
      {
        /// Send it to the transport and hope remote handles fragmented UDP
        if( msg.IsRequest() )
        {
          SIPURI ruri;
          msg.GetRequestURI( ruri );
          ruri.SetParameter( "transport", "udp" );
          msg.SetRequestURI( ruri );
        }

        m_Manager.OnSendMessageToTransport( msg, *this );
        return TRUE;
      }

      /// now pop the top via and replace it with this sockets address
      SIPParser::Via via = msg.PopTopVia();
      PIPSocket::Address addr;
      WORD port;
      dynamic_cast<PTCPSocket*>(m_Socket)->GetLocalAddress( addr, port );
      via.SetPort( port );
      via.SetProtocol( "TCP" );
      msg.AppendVia( via );

      /// change the via of the opening request as well
      m_OpeningRequest.PopTopVia();
      m_OpeningRequest.AppendVia( via );

      if( msg.GetContactSize() > 0 && !msg.IsRegister() )
      {
        SIPURI contactURI = msg.GetContactTopURI();
        if( m_Manager.IsLocalAddressAndPort( contactURI ) )
        {
          /// we need to change the contact as well
          //SIPParser::Contact contact = SIPMessage::GetContactAt(0);
          contactURI = via.GetURI();
          contactURI.SetParameter( "transport", "tcp" );
          msg.SetTopContactURI( contactURI );
          m_OpeningRequest.SetTopContactURI( contactURI );
        }
      }
      packet = msg.AsString();
    }else if( m_Socket == NULL && (m_Type == NIST || m_Type == IST) )
    {
      /// oops! A large response for a request sent over UDP!
      // Send it to the transport and hope remote handles fragmented UDP
      m_Manager.OnSendMessageToTransport( msg, *this );
        return TRUE;
    }

    if( !m_Socket->IsOpen() )
    {
      PTCPSocket * s = CreateTCPSocket( m_SocketTargetAddress );
      if( s == NULL )       
        return FALSE;

      delete m_Socket;
      m_Socket = s;
    }
  }

  msg.SetInternalHeader( "SEND-EVENT", "TRUE" );
  PIPSocket::Address sendAddr;
  WORD sendPort;
  static_cast<PTCPSocket*>(m_Socket)->GetPeerAddress( sendAddr, sendPort );

  OStringStream traceStream;
  traceStream <<  "SEND: enc=" << msg.IsEncrypted() 
    << " " <<  packet.GetLength() << " Bytes to " 
    << sendAddr << ":" << sendPort << ":TCP (" 
    << msg.GetStartLine() << ")" 
    << " Interface Address=(*)"
    << "\r\n" << packet;

  LOG_CONTEXT_IF_DEBUG( LogInfo(), msg.GetCallId(), traceStream );

  PTRACE( 1, traceStream.str() );

  SIPTransport::NotifyWrite( traceStream.str() );

  if( m_Socket->Write( packet.GetPointer(), packet.GetLength() ) )
  {
    EnqueueEvent( new SIPEvent( new SIPMessage(msg) ) );
    return TRUE;
  }

  return FALSE;
#else
  m_Manager.OnSendMessageToTransport( msg, *this );
  return TRUE;
#endif

}

BOOL SIPTransaction::DoDNSFailover()
{
   /* 
   For SIP requests, failure occurs if the transaction layer reports a
   503 error response or a transport failure of some sort (generally,
   due to fatal ICMP errors in UDP or connection failures in TCP).
   Failure also occurs if the transaction layer times out without ever
   having received any response, provisional or final (i.e., timer B or
   timer F in RFC 3261 [1] fires).  If a failure occurs, the client
   SHOULD create a new request, which is identical to the previous, but
   has a different value of the Via branch ID than the previous (and
   therefore constitutes a new SIP transaction).  That request is sent
   to the next element in the list as specified by RFC 2782.
   */

  SIPMessage request = m_OpeningRequest;
  request.ClearInternalHeaders();

  if( m_FailOverAttempts == 3 || !request.IsValid() || !request.IsRequest()  )
    return FALSE;

  m_FailOverAttempts++;

  SIPSrvRecord srv( request );
  if( srv.GetSize( "UDP" ) <= m_FailOverAttempts )
    return FALSE;

  const SIPSrvRecord::HostRecord * hr = srv.GetRecordAt( m_FailOverAttempts, "UDP" );

  if( !hr->m_Address.IsValid() )
    return FALSE;

  SIPURI failOverURI( hr->m_Address, hr->m_Port );
  
  /// check if it will loopback
  if( m_Manager.IsLocalAddressAndPort( failOverURI ) )
    return FALSE;

  request.SetSendAddress( failOverURI );

  /// generate a new Via
  Via via = request.PopTopVia();
  OString newBranch = via.GetBranch() + "-";
  via.SetBranch( newBranch );
  request.AppendVia( via );
  TransactionId trnId;

  m_Manager.OnDNSFailOver( request, *this );
  return m_Manager.FindTransactionAndAddEvent( request, trnId, FALSE, m_FailOverAttempts ); 
}




